#include <QApplication>
#include <QGroupBox>
#include <QStringList>
#include <QCheckBox>
#include <QPushButton>
#include <QLabel>
#include <QHBoxLayout>
#include <QDebug>
#ifdef QT_DIAG_LIB
# include <eventfilter.h>
# include <qwidgetdump.h>
#endif
#if QT_VERSION >= 0x050000
# include <QWindow>
# include <QScreen>
#endif

const Qt::WindowFlags frameLessFlags = Qt::Window | Qt::CustomizeWindowHint
    | Qt::FramelessWindowHint;

const Qt::WindowFlags defaultFlags = Qt::Window| Qt::WindowTitleHint
    | Qt::WindowSystemMenuHint | Qt::WindowMinMaxButtonsHint
    | Qt::WindowCloseButtonHint
#if QT_VERSION >= 0x050000
    | Qt::WindowFullscreenButtonHint
#endif
    ;

QString scalingDescription(const QWidget *w)
{
    QString r;
    QDebug str(&r);
#if QT_VERSION >= 0x050000
    str.noquote();
#endif
    str.nospace();
#if QT_VERSION >= 0x050000
    QScreen *screen = 0;
    if (const QWidget *topL = w->nativeParentWidget()) {
        if (QWindow *w = topL->windowHandle())
            screen = w->screen();
    }
    str << screen << ", devicePixelRatio=" << w->devicePixelRatio();
#endif
    str << ", resolution=" << w->logicalDpiX() << ','
        << w->logicalDpiY() << "dpi";
    return r;
}

class ScreenTrackerLabel : public QLabel
{
    Q_OBJECT
public:
    ScreenTrackerLabel(QWidget *parent = 0) :
        QLabel(parent)
    {
        updateText();
        setWordWrap(true);
        setTextInteractionFlags(Qt::TextBrowserInteraction);
    }

public slots:
    void updateText() { setText(scalingDescription(this)); }
};

class Widget : public QGroupBox
{
    Q_OBJECT
public:
    Widget(QString t, Qt::WindowFlags flags) : QGroupBox(t)
    {
        setWindowFlags(flags);
        setObjectName("MainW");
        setWindowTitle(t);
        QHBoxLayout *mainL = new QHBoxLayout(this);
        QVBoxLayout *vl = new QVBoxLayout;
        QCheckBox *cb = new QCheckBox("Frameless", this);
        cb->setChecked(isFrameLess());
        cb->setMinimumWidth(200);
        connect(cb, SIGNAL(toggled(bool)),this,SLOT(toggleFrameLess(bool)));
        vl->addWidget(cb);
        QPushButton *qB = new QPushButton("Quit", this);
        QPushButton *dumpB = new QPushButton("Dump", this);
        vl->addWidget(dumpB);
        connect(dumpB, SIGNAL(clicked()), this, SLOT(dumpWidgets()));
        dumpB->setShortcut(Qt::CTRL + Qt::Key_D);
        vl->addWidget(qB);
        connect(qB, SIGNAL(clicked()), this, SLOT(close()));
        qB->setShortcut(Qt::CTRL + Qt::Key_Q);

        m_screenTrackerLabel = new ScreenTrackerLabel(this);
        vl->addWidget(m_screenTrackerLabel);
        vl->addItem(new QSpacerItem(0, 0, QSizePolicy::Ignored, QSizePolicy::MinimumExpanding));

        mainL->addLayout(vl);
        mainL->addItem(new QSpacerItem(0, 0, QSizePolicy::MinimumExpanding, QSizePolicy::Ignored));
    }

    bool isFrameLess() const { return windowFlags() & Qt::FramelessWindowHint; }

public slots:
    void toggleFrameLess(bool frameLess)
    {
        static int n = 0;
        Qt::WindowFlags newFlags = frameLess ? frameLessFlags : defaultFlags;
        qDebug() << "###" << n << ">setWindowFlags(" << newFlags << ')';
        setWindowFlags(newFlags);
        qDebug() << "###" << n << "<setWindowFlags()/>show()";
        show();
        qDebug() << "###" << n << "<show()";
        n++;
    }

    void dumpWidgets()
    {
#ifdef QT_DIAG_LIB
        QtDiag::dumpAllWidgets();
#endif
    }

    void updateScreenTrackerLabel() { m_screenTrackerLabel->updateText(); }

private:
    ScreenTrackerLabel *m_screenTrackerLabel;
};

int main(int argc, char **argv)
{
    QStringList args;
    std::copy(argv + 1, argv + argc, std::back_inserter(args));

#if QT_VERSION >= 0x050600
    if (args.contains("-n")) {
        qDebug() << "DisableHighDpiScaling";
        QCoreApplication::setAttribute(Qt::AA_DisableHighDpiScaling);
    } else if (args.contains("-s")) {
        qDebug() << "EnableHighDpiScaling";
        QCoreApplication::setAttribute(Qt::AA_EnableHighDpiScaling);
    }
#endif

    QApplication app( argc, argv );

    QString t = QT_VERSION_STR;
#if QT_VERSION >= 0x050000
    t += ' '  + QGuiApplication::platformName();
#endif

    qDebug() << t << args;

#ifdef QT_DIAG_LIB
    if (args.contains("-e")) {
        QtDiag::EventFilter::EventCategories cats = QtDiag::EventFilter::GeometryEvents
            | QtDiag::EventFilter::StateChangeEvents;
        app.installEventFilter(new QtDiag::EventFilter(cats, &app));
    }
#endif

    const Qt::WindowFlags flags = args.contains("-f") ? frameLessFlags : defaultFlags;
    Widget mainWindow(t, flags);
    mainWindow.setWindowTitle(t);

    mainWindow.show();
#if QT_VERSION >= 0x050000
    mainWindow.updateScreenTrackerLabel();
    QObject::connect(mainWindow.windowHandle(), &QWindow::screenChanged,
                     &mainWindow, &Widget::updateScreenTrackerLabel);
#endif
    return app.exec();
}

#include "main.moc"
